'''
dict_xpath: 一种用于"快速定位Python字典的key"的数据结构.

该数据结构线程安全.

当对象的数据结构具有 不确定性、易变性, 或者要获取的值处于多层嵌套的较深层时, 可使用该函数, 示例:
    nums = {
        '1': 1,
        '2': {
            '21': 21, '22': {
                '221': 221, '222': 222, '223': 223, '99': 2099
            },
            '23': 23
        },
        '3': {
            '31': 31, '32': {
                '321': 321, '322': 322, '323': 323, '99': 3099
            },
            '33': 33
        },
        '4': 4
    }

    s = dict_xpath(nums)

    s.xpath(path='33') -> 33
    s.xpath(path='2 > 99') -> 2099
    s.xpath(path='3 > 99') -> 3099

注意: path无须写出路径中的所有元素, 只要写出的元素能够确定一条唯一的路径即可.
'''


import re


class dict_xpath():

    def __init__(self, item):
        self.__item = item
    
    def __brisk(self, item, i, total, path):
        if (i < total) and (type(item) is dict):
            ki = path[i]
            for k,v in item.items():
                if k == ki:
                    if i == total - 1: return v, True
                    else:
                        value, r = self.__brisk(v, i + 1, total, path)
                        if r: return value, True
                value, r = self.__brisk(v, i, total, path)
                if r: return value, True
        return None, False
    
    def xpath(self, path, default=Exception("key does not exist")):
        path = re.split(' *> *', path)
        value, r = self.__brisk(self.__item, 0, len(path), path)
        if r:
            return value
        elif type(default) is Exception:
            raise default
        else:
            return default
