'''
laydb: 一种类文件夹的数据库

@author: jutooy

该数据结构线程安全.
'''


from pickle import loads as pickle_loads
from json import loads as json_loads
from .localtool import write_pickle, write_json
from .localtool import set_path, libpath


class MYKEYERR(): ...


class laydb():
    
    def __init__(self, folder, dump_mode='pickle'):
        self.i = item = {}
        self.s = sons = {}
        self.folder = folder
        self.item_folder = folder + '/__item__/'
        self.dump_mode = dump_mode

        self.fetch = item.__getitem__
        self.get = item.get
        self.items = item.items
        self.keys = item.keys
        self.values = item.values

        self.sonkeys = sons.keys
        self.sonvalues = sons.values
        self.sonitems = sons.items

        if dump_mode == 'pickle':
            self.dump = write_pickle
        elif dump_mode == 'json':
            self.dump = write_json
        set_path(self.item_folder)
        self.load_all()

    def load_all(self):
        i = self.i
        s = self.s
        set_path(self.folder)
        if self.dump_mode == 'pickle':
            for child in libpath(self.folder).iterdir():
                if child.is_dir():
                    name = child.name
                    if name[:7] == '__son__':
                        s[name[7:]] = laydb(folder=str(child), dump_mode='pickle')
                    elif name[:8] == '__item__':
                        for ch in child.iterdir():
                            if ch.suffix == '.lay_pickle':
                                i[ch.name[:-11]] = pickle_loads(ch.read_bytes())
        elif self.dump_mode == 'json':
            for child in libpath(self.folder).iterdir():
                if child.is_dir():
                    name = child.name
                    if name[:7] == '__son__':
                        s[name[7:]] = laydb(folder=str(child), dump_mode='json')
                    elif name[:8] == '__item__':
                        for ch in child.iterdir():
                            if ch.suffix == '.lay_json':
                                i[ch.name[:-9]] = json_loads(ch.read_text(encoding='utf8'))
    
    # 针对 self.i
    def set(self, key, value):
        fp = self.item_folder + key + '.lay_' + self.dump_mode  # 不要使用 f"", 因为要验证key是否为字符串
        self.i[key] = value
        self.dump(fp, self.i[key])  # 使用 self.i[key] 而非 value, 使写入硬盘的数据总是与self.i[key]相同
        return True

    # 针对 self.i
    def setdefault(self, key, default=MYKEYERR):
        try:
            return self.i[key]
        except Exception as e:
            if default is MYKEYERR:
                raise e
            else:
                self.set(key, default)
                return self.i[key]  # except里的self.i[key]可以视为try里的self.i[key], 即: 视为"未发生异常, 未进入except"

    # 针对 self.s
    def __getitem__(self, key):
        son = self.s.get(key)
        if not son:
            son = self.s.setdefault(key, laydb(folder=self.folder + '/__son__' + key,  # 不要使用 f""
                dump_mode=self.dump_mode))
        return son

    # 全数据视图
    def view(self):
        r = {}
        r['i'] = self.i
        r['s'] = {skey:son.view() for skey, son in list(self.s.items())}
        return r
