'''
Created on Sat Oct 30 07:25:31 2021
@author: jutooy

laydict: 一种类文件夹的数据结构.

该数据结构线程安全.


一个 "laydict" 实例就是一个"文件夹", 每个"文件夹"内部维护两个容器, 分别是 "self.i" 和 "self.s" .

"self.i" 以"键值对"的形式存放"文件", "self.s" 以"键值对"的形式存放"子文件夹".

每个"子文件夹"在结构上和"父文件夹"一致, 维护着各自的 "self.i" 和 "self.s" .

文件夹的层次数没有限制, 只要你愿意, 可以把一个"文件"存放在第100万层、1000万层、1亿层...深度的文件夹中.
'''


class laydict():
    
    def __init__(self):
        self.i = item = {}
        self.s = sons = {}

        # 针对 self.i 的方法:
        self.set = item.__setitem__
        self.setmany = item.update
        self.fetch = item.__getitem__
        self.get = item.get
        self.setdefault = item.setdefault
        self.items = item.items
        self.keys = item.keys
        self.values = item.values
        self.pop = item.pop
        self.popitem = item.popitem

        # 针对 self.s 的方法:
        self.sonkeys = sons.keys
        self.sonvalues = sons.values
        self.sonitems = sons.items

    # 针对 self.s
    # setson
    def __setitem__(self, key, value):
        if type(value) is laydict:
            self.s[key] = value
            return True
        else:
            raise Exception('The value must be an instance of laydict')
    
    # 针对 self.s
    # getson
    def __getitem__(self, key):
        son = self.s.get(key)
        if not son:
            son = self.s.setdefault(key, laydict())
        return son

    # 针对 self.s
    def popson(self, key):
        return self.s.pop(key, None) or laydict()
        # 不要用 [return self.s.pop(key, laydict())] 这种方式, 因为 laydict.__init__ 有较多的执行操作.

    # 全数据视图
    def view(self):
        r = {}
        r['i'] = self.i
        r['s'] = {skey:son.view() for skey, son in list(self.s.items())}
        return r
