'''
limdict : 一种限定容量的有序字典
该数据类型线程安全.
'''


from collections import OrderedDict
from .varlock import varlock


class MYKEYERR(): ...


class limdict():

    def __init__(self, volume=None):
        self.__volume = volume
        self.__d = OrderedDict()
        self.__lock = varlock()

    async def set_volume(self, volume):
        async with self.__lock:
            self.__volume = volume

    async def add(self, key, value):
        expired = []
        d = self.__d
        async with self.__lock:
            d.pop(key, None)
            d[key] = value
            if self.__volume:
                while len(d) > self.__volume:
                    expired.append(d.popitem(last=False))
        return expired

    async def fetch(self, key):
        async with self.__lock:
            return self.__d[key]

    async def get(self, key, default=None):
        async with self.__lock:
            return self.__d.get(key, default)

    async def pop(self, key, default=MYKEYERR):
        async with self.__lock:
            if default is MYKEYERR:
                return self.__d.pop(key)
            else:
                return self.__d.pop(key, default)
    
    async def keys(self):
        async with self.__lock:
            return list(self.__d.keys())
    
    async def values(self):
        async with self.__lock:
            return list(self.__d.values())

    async def items(self):
        async with self.__lock:
            return list(self.__d.items())
