'''
limset : 一种限定容量的有序集合
该数据类型线程安全.
'''


from collections import OrderedDict
from .varlock import varlock


class limset():

    def __init__(self, volume=None):
        self.__volume = volume
        self.__d = OrderedDict()
        self.__lock = varlock()

    async def set_volume(self, volume):
        async with self.__lock:
            self.__volume = volume

    async def add(self, value):
        expired = []
        d = self.__d
        async with self.__lock:
            d.pop(value, None)
            d[value] = None
            if self.__volume:
                while len(d) > self.__volume:
                    expired.append(d.popitem(last=False)[0])
        return expired

    async def add_many(self, values):
        expired = []
        d = self.__d
        async with self.__lock:
            for value in values:
                d.pop(value, None)
                d[value] = None
            if self.__volume:
                while len(d) > self.__volume:
                    expired.append(d.popitem(last=False)[0])
        return expired

    async def remove(self, value, unraise=True):
        async with self.__lock:
            if unraise:
                self.__d.pop(value, None)
            else:
                self.__d.pop(value)
        return True
    
    async def remove_many(self, values, unraise=True):
        d = self.__d
        async with self.__lock:
            if unraise:
                for value in values: d.pop(value, None)
            else:
                for value in values: d.pop(value)
        return True

    async def values(self):
        async with self.__lock:
            return list(self.__d.keys())
