import hashlib, re
from datetime import datetime
from math import ceil
from json import dumps as jdumps
from json import loads as jloads
from pickle import dumps as pdumps
from pickle import loads as ploads
''' 允许其它模块导入 start '''
from pathlib import Path as libpath
from os.path import join as path_join
from os.path import split as path_split
from os.path import abspath as path_abs
from os.path import exists as path_exists
from base64 import encodebytes as bytes_into_base64
from base64 import b64decode as base64_into_bytes
''' 允许其它模块导入 end '''


class MYKEYERR(): ...

def prepath(path):  # 把路径转化成 /a/b/c/ 的格式
    path = path_abs(path).replace('\\', '/')
    if path[-1:] == '/':
        return path
    else:
        return path + '/'

# -> '2021-10-24'
nowdate = get_nowdate = lambda : str(datetime.now())[:10]

# -> '2021-10-24 20:19:10'
nowtime = get_nowtime = lambda : str(datetime.now())[:19]

# -> '[2021-10-24 20:18:57]'
pretty_time = get_pretty_time = lambda : f"[{get_nowtime()}]"


def get_md5(text):
    return hashlib.md5(text.encode(encoding='utf-8')).hexdigest()


def cut_data(data, size):
    return [data[size*(i-1): size*i] for i in range(1, ceil(len(data)/size)+1)]


def set_path(path):
    libpath(path).mkdir(parents=True, exist_ok=True)

def write_json(fpath, data, ensure_ascii=False):
    data = jdumps(data, ensure_ascii=ensure_ascii)
    libpath(fpath).write_text(data, encoding='utf8')

def read_json(fpath, default=MYKEYERR):
    try:
        return jloads(libpath(fpath).read_text(encoding='utf8'))
    except Exception as e:
        if default is MYKEYERR:
            raise e
        else:
            return default

def write_pickle(fpath, data):
    libpath(fpath).write_bytes(pdumps(data))

def read_pickle(fpath, default=MYKEYERR):
    try:
        return ploads(libpath(fpath).read_bytes())
    except Exception as e:
        if default is MYKEYERR:
            raise e
        else:
            return default

def check_chinese(text, pattern):
    '''
    text_check_chinese('abc一二三', 'any') -> False
    text_check_chinese('一二三', 'any') -> True
    text_check_chinese('abc一二三', 'all') -> False
    text_check_chinese('一二三', 'all') -> True
    '''
    if pattern == 'any':
        return bool(re.search('[\u4e00-\u9fa5]', text))
    elif pattern == 'all':
        return not bool(re.search('[^\u4e00-\u9fa5]', text))
    else:
        raise Exception('pattern err')

def func_get_name(func):
    '''
    def test(): pass
    func_get_name(test) -> 'test'
    '''
    name = str(func)
    name = re.split(' of | at |>', name)[0]
    name = re.sub("^<function |^<bound method |^<class |<| object|'", '', name)
    return name

def obj_get_name(obj):
    ''' obj_get_name('a') -> 'class_str_object' '''
    typ = str(type(obj))
    if type(obj) in (int, list, str, dict, set, tuple, float, bool):
        name = re.findall("'(.+)'", typ)[0]
        return f"class_{name}_object"
    else:
        return func_get_name(obj)

def limit_input(prompt='', limit=None):
    if type(limit) in (list, tuple, set, dict, str):
        while True:
            user = input(prompt)
            if user in limit: return user
    elif limit is int:
        while True:
            try:
                return int(input(prompt))
            except: pass
    elif limit is float:
        while True:
            try:
                return float(input(prompt))
            except: pass
    else:
        return input(prompt)

a_z = 'abcdefghijklmnopqrstuvwxyz'
A_Z = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
