'''
Python上最优雅的MongoDB ORM

@作者: jutooy
@邮箱: jutooy@qq.com

基于 pymongo--3.12.1 开发
'''


from copy import deepcopy
from pymongo import MongoClient  # from pyzing.mongodb import mongo, mc, mo, MongoClient


class Oper():...


class re(Oper):
    def __init__(self, s, i=False):
        if i:
            self.v = {'$regex': s, '$options': 'i'}
        else:
            self.v = {'$regex': s}


class contain(Oper):
    # 针对 array column, 指该column是否包含某个元素
    def __init__(self, o):
        if type(o) in (list, set, tuple, dict):  # 可接收容器
            o = list(set(o))
            if len(o) == 0: raise Exception("Empty containers are not supported")
            elif len(o) == 1: self.v = {'$elemMatch':{'$eq':o[0]}}
            else: self.v = [{'$elemMatch':{'$eq':x}} for x in o]
        else:  # 可接收单个元素
            self.v = {'$elemMatch':{'$eq':o}}


class uncontain(Oper):
    def __init__(self, o):
        if type(o) in (list, set, tuple, dict):  # 可接收容器
            o = list(set(o))
            if o: self.v = [{'$elemMatch':{'$eq':x}} for x in o]
            else: raise Exception("Empty containers are not supported")
        else:  # 可接收单个元素
            self.v = [{'$elemMatch':{'$eq':o}}]


class isin(Oper):
    def __init__(self, *v):
        len_ = len(v)
        if len_ > 1: self.v = list(set(v))
        elif len_ == 1:
            v = v[0]
            try:
                self.v = list(set(v))
            except:
                self.v = [v]
        else: self.v = []
        self.v = {'$in': self.v}


class notin(Oper):
    def __init__(self, *v):
        len_ = len(v)
        if len_ > 1: self.v = list(set(v))
        elif len_ == 1:
            v = v[0]
            try:
                self.v = list(set(v))
            except:
                self.v = [v]
        else: self.v = []
        self.v = {'$nin': self.v}


class Filter():

    def __init__(self, name=None, where=None):
        self.__name = name or ''
        self.__where = where or {}

    def __getattr__(self, o):
        if o == 'a4173c5c8a87db003':
            return self.__where
        elif self.__name:
            return Filter(name=f"{self.__name}.{o}")
        else:
            return Filter(name=o)

    def __ne__(self, o):  # !=
        return Filter(name=self.__name, where={self.__name: {'$ne': o}})
    
    def __lt__(self, o):  # <
        return Filter(name=self.__name, where={self.__name: {'$lt': o}})

    def __le__(self, o):  # <=
        return Filter(name=self.__name, where={self.__name: {'$lte': o}})
    
    def __gt__(self, o):  # >
        return Filter(name=self.__name, where={self.__name: {'$gt': o}})
    
    def __ge__(self, o):  # >=
        return Filter(name=self.__name, where={self.__name: {'$gte': o}})
    
    def __and__(self, o):  # &
        s, o = deepcopy(self.__where), deepcopy(o.__where)
        where = {'$and': [s, o]} if (set(s) & set(o)) else {**s, **o}
        return Filter(name='', where=where)
    
    def __or__(self, o):  # |
        s, o = deepcopy(self.__where), deepcopy(o.__where)
        where = {'$or': [s, o]}
        return Filter(name='', where=where)
    
    # def __invert__(self): ...  # ~
    # not的语法太死板, 没有太大意义, 故不予支持.

    def __eq__(self, o):  # ==
        if isinstance(o, uncontain):
            where = {'$nor': [{self.__name: x} for x in deepcopy(o.v)]}
        elif isinstance(o, Oper):
            if isinstance(o, contain) and (type(o.v) is list):
                where = {'$and': [{self.__name: x} for x in deepcopy(o.v)]}
            else:
                where = {self.__name: deepcopy(o.v)}
        else:
            where = {self.__name: o}
        return Filter(name=self.__name, where=where)


class mongo():

    def __init__(self, sheet, where=None, column=None):
        self.sheet = sheet
        self.where = where or {}
        self.column = column or set()
        self.insert = sheet.insert_one
        self.insert_many = sheet.insert_many

    def __add__(self, data):
        if type(data) is dict:
            r = self.insert(data)
            return r.acknowledged, r.inserted_id  # True, _id
        else:
            r = self.insert_many(data)
            return r.acknowledged, r.inserted_ids

    # 对某一个 array column添加元素
    # 当某条数据的该column已存在相同元素时, 该column不再添加(若要添加, 可尝试用$push替换$addToSet).
    def add_for_array(self, column, o):
        where = deepcopy(self.where)
        r = self.sheet.update_many(where, {'$addToSet': {column: o}})
        return r.acknowledged, r.matched_count

    def copy(self, where=None, column=None):
        if where is None: where = deepcopy(self.where)
        if column is None: column = deepcopy(self.column)
        return mongo(sheet=self.sheet, where=where, column=column)

    def clear_column(self): return self.copy(column=set())
    
    def clear_condition(self): return self.copy(where={})

    def __getitem__(self, key):
        if type(key) is int: key = slice(key, {}, None)
        type_ = type(key)
        # 切片取值
        if type_ is slice:
            start, B, C = key.start, key.stop, key.step
            if type(start) is float: raise Exception("Float type is not supported")
            if type(start) is int and start < 0: raise Exception("Slice must be >= 0")
            start = start or 0
            if type(B) is float: raise Exception("Float type is not supported")
            if type(B) is int:
                if B < 0: raise Exception("Slice must be >= 0")
                if B <= start: return []
            if type(B) is dict:
                type_data = 'kw'
                size = 1
                rename = B.get('rename') or {}
                sort = B.get('sort') or []
            else:
                C = C or {}
                type_data = 'list'
                size = B - start if B else None
                rename = C.get('rename') or {}
                sort = C.get('sort') or []
            if len(rename) != len(set(rename.values())): raise Exception('Duplicate new name')
            sel = self.column
            if sel:
                column = dict.fromkeys(sel, 1)
                column.setdefault('_id', 0)
            else:
                column = None  # 只能是None, 不能是空字典{}, 否则会只返回 _id
            sh = self.sheet.find(deepcopy(self.where), column)
            # sort = [(作者, True), (字数, True), ...], True表示升序
            if sort: sh = sh.sort([(x, 1 if y else -1) for x, y in sort])
            if start: sh = sh.skip(start)
            if size: sh = sh.limit(size)
            kws = list(sh)
            if sel: kws = [{key: kw.get(key) for key in sel} for kw in kws]
            if rename:
                for kw in kws:
                    data = {new: kw.pop(old) for old,new in rename.items() if old in kw}
                    kw.update(data)
            if type_data == 'kw': kws = kws[0]
            return kws
        # 限定column
        elif type_ in (str, tuple):  # 输入多个字符串, 用逗号隔开, Python会自动打包成tuple
            column = {key} if (type_ is str) else set(key)
            return self.copy(column=column)
        # Filter
        elif type_ is Filter:
            o = key.a4173c5c8a87db003
            if self.where:
                s = self.where
                where = {'$and': [s, o]} if (set(s) & set(o)) else {**s, **o}
            else:
                where = o
            return self.copy(where=deepcopy(where))
        else:
            return None

    def delete_one(self):
        r = self.sheet.delete_one(deepcopy(self.where))
        return r.acknowledged, r.deleted_count

    def delete_all(self):
        r = self.sheet.delete_many(deepcopy(self.where))
        return r.acknowledged, r.deleted_count

    def reset_one(self, kw):
        kw = deepcopy(kw or {})
        if self.column:
            kw = {k: kw.get(k) for k in self.column}
            r = self.sheet.update_one(deepcopy(self.where), {'$set':kw})
            return r.acknowledged, r.matched_count
        else:
            raise Exception('Column must be specified')

    def reset_all(self, kw):
        kw = deepcopy(kw or {})
        if self.column:
            kw = {k: kw.get(k) for k in self.column}
            r = self.sheet.update_many(deepcopy(self.where), {'$set':kw})
            return r.acknowledged, r.matched_count
        else:
            raise Exception('Column must be specified')

    def update_one(self, kw):
        kw = deepcopy(kw or {})
        if self.column: kw = {k:v for k,v in kw.items() if k in self.column}
        if kw:
            r = self.sheet.update_one(deepcopy(self.where), {'$set':kw})
            return r.acknowledged, r.matched_count
        else:
            return True, 0

    def update_all(self, kw):
        kw = deepcopy(kw or {})
        if self.column: kw = {k:v for k,v in kw.items() if k in self.column}
        if kw:
            r = self.sheet.update_many(deepcopy(self.where), {'$set':kw})
            return r.acknowledged, r.matched_count
        else:
            return True, 0
    # matched_count 与 modified_count 的区别:
    # matched_count 表示匹配到的数目, 如果是update_one, 则 matched_count in [0, 1]
    # modified_count 表示数据有变化的数目
    # 如果一条数据修改前和修改后一致(例如:把3修改成3), 则不会统计到modified_count中


mc = Filter()  # 表示: mongodb_column


class mo():  # 表示: mongodb_oper
    re = re
    contain = contain
    uncontain = uncontain
    isin = isin
    notin = notin
