'''
Python上最优雅的mysql ORM

@作者: jutooy
@邮箱: jutooy@qq.com

基于 pymysql--1.0.2 开发
'''


import asyncio
from copy import deepcopy
from collections import deque
from json import dumps as json_dumps
from pymysql import connect
from pymysql.cursors import DictCursor


jdumps = lambda x: json_dumps(x, ensure_ascii=False)


class Oper():...

class re(Oper):
    def __init__(self, s):
        if s:
            self.v = f"regexp {jdumps(s)}"
        else:
            raise Exception("Regular expression content cannot be empty")

class isin(Oper):
    def __init__(self, *v):
        len_ = len(v)
        if len_ > 1: self.v = list(set(v))
        elif len_ == 1:
            v = v[0]
            try:
                self.v = list(set(v))
            except:
                self.v = [v]
        else: self.v = []

class notin(Oper):
    def __init__(self, *v):
        len_ = len(v)
        if len_ > 1: self.v = list(set(v))
        elif len_ == 1:
            v = v[0]
            try:
                self.v = list(set(v))
            except:
                self.v = [v]
        else: self.v = []


def parser_isin(name, o):
    if not o: return f"({name} is null) and ({name} is not null)"
    o = set(o)
    condition_null = None
    if None in o:
        condition_null = f"{name} is null"
        o.remove(None)
        if not o: return condition_null
    types = set(type(x) for x in o)
    if (types == {str}) or (types == {int}):
        condition_in = f"{name} in ({','.join(jdumps(x) for x in o)})"
        if condition_null: return f"({condition_null}) or ({condition_in})"
        return condition_in
    raise Exception("isin only receives [None, int] or [None, str]")

def parser_notin(name, o):
    o = set(o)
    if not o: return f"({name} is null) or ({name} is not null)"
    elif None in o:
        o.remove(None)
        condition_null = f"{name} is not null"
        if o:
            types = set(type(x) for x in o)
            if (types != {str}) and (types != {int}): raise Exception("notin only receives [None, int] or [None, str]")
            condition_in = f"{name} not in ({','.join(jdumps(x) for x in o)})"
            return f"({condition_null}) and ({condition_in})"
        return condition_null
    else:
        types = set(type(x) for x in o)
        if (types != {str}) and (types != {int}): raise Exception("notin only receives [None, int] or [None, str]")
        condition_in = f"{name} not in ({','.join(jdumps(x) for x in o)})"
        condition_null = f"{name} is null"
        return f"({condition_null}) or ({condition_in})"


class Filter():

    def __init__(self, name='', where=''):
        self.__name = name or ''
        self.__where = where or ''
    
    def __getattr__(self, o):
        if o == 'a4173c5c8a87db003':
            return self.__where
        else:
            return Filter(name=o)
    
    def __ne__(self, o):  # !=
        if o is None:
            return Filter(name=self.__name, where=f"{self.__name} is not null")
        else:
            return Filter(name=self.__name, where=f"{self.__name} != {jdumps(o)}")
    
    def __lt__(self, o):  # <
        return Filter(name=self.__name, where=f"{self.__name} < {jdumps(o)}")

    def __le__(self, o):  # <=
        return Filter(name=self.__name, where=f"{self.__name} <= {jdumps(o)}")

    def __gt__(self, o):  # >
        return Filter(name=self.__name, where=f"{self.__name} > {jdumps(o)}")
    
    def __ge__(self, o):  # >=
        return Filter(name=self.__name, where=f"{self.__name} >= {jdumps(o)}")

    def __and__(self, o):  # &
        s, o = self.__where, o.__where
        if s == o:
            return Filter(name='', where=s)
        else:
            return Filter(name='', where=f"({s}) and ({o})")
    
    def __or__(self, o):  # |
        s, o = self.__where, o.__where
        if s == o:
            return Filter(name='', where=s)
        else:
            return Filter(name='', where=f"({s}) or ({o})")

    def __invert__(self):  # ~
        return Filter(name='', where=f"not ({self.__where})")

    def __eq__(self, o):  # ==
        typ = type(o)
        name = self.__name
        if typ is re:
            return Filter(name=name, where=f"{name} {o.v}")
        elif typ is isin:
            return Filter(name=name, where=parser_isin(name, o.v))
        elif typ is notin:
            return Filter(name=name, where=parser_notin(name, o.v))
        elif o is None:
            return Filter(name=name, where=f"{name} is null")
        else:
            return Filter(name=name, where=f"{name} = {jdumps(o)}")


class msheet():

    def __init__(self, address, sheet, where=None, column=None, conn_pool=None, active=False):
        self.address = address
        self.sheet = sheet
        self.where = where or ''
        self.column = column or set()

        # self.conn_pool 将在 class mpool() 中被调用
        # 禁止: self.conn_pool = conn_pool or [], 因为传入的有可能是一个空容器
        self.conn_pool = conn_pool

        if active:
            self.mkconn()
        else:
            self.conn = None

    def mkconn(self):
        self.conn = connect(**self.address)
    
    def __get_conn(self):
        if self.conn is None:
            self.mkconn()
        else:
            self.conn.ping(reconnect=True)
        cursor = self.conn.cursor(cursor=DictCursor)
        return self.conn, cursor, self.sheet

    def __copy(self, where=None, column=None):
        if where is None:
            where = self.where
        if column is None:
            column = deepcopy(self.column)
        orm = msheet(address=self.address, sheet=self.sheet, where=where, column=column)
        orm.conn = self.conn
        return orm

    def copy(self):
        return msheet(
            address = self.address,
            sheet = self.sheet,
            where = self.where,
            column = deepcopy(self.column)
        )

    def __add__(self, data):
        conn, cursor, sheet = self.__get_conn()
        if type(data) is dict:
            字段s = data.keys()
            sql = f"insert into {sheet}({', '.join(字段s)}) VALUES ({', '.join(('%s',)*len(字段s))})"
            try:
                cursor.execute(sql, tuple(data.values()))
                conn.commit()
                cursor.close()
                return cursor  # cursor.rowcount, cursor.lastrowid
            except Exception as err:
                conn.rollback()
                raise err
        else:
            cols = set()
            for x in data: cols |= set(x)
            cols = list(cols)
            sql = f"insert into {sheet}({', '.join(cols)}) VALUES ({', '.join(('%s',)*len(cols))})"
            datas = tuple(tuple(x.get(k) for k in cols) for x in data)
            try:
                cursor.executemany(sql, datas)
                conn.commit()
                cursor.close()
                return cursor
            except Exception as err:
                conn.rollback()
                raise err
    
    def clear_column(self): return self.__copy(column=set())
    
    def clear_condition(self): return self.__copy(where='')

    def __getitem__(self, key):
        if type(key) is int: key = slice(key, {}, None)
        type_ = type(key)
        # 切片取值
        if type_ is slice:
            start, B, C = key.start, key.stop, key.step
            if type(start) is float: raise Exception("Float type is not supported")
            if type(start) is int and start < 0: raise Exception("Slice must be >= 0")
            start = start or 0
            if type(B) is float: raise Exception("Float type is not supported")
            if type(B) is int:
                if B < 0: raise Exception("Slice must be >= 0")
                if B <= start: return []
            if type(B) is dict:
                type_data = 'kw'
                size = 1
                rename = B.get('rename') or {}
                sort = B.get('sort') or []
            else:
                C = C or {}
                type_data = 'list'
                size = B - start if B else None
                rename = C.get('rename') or {}
                sort = C.get('sort') or []
            if len(rename) != len(set(rename.values())): raise Exception('Duplicate new name')
            sel = ', '.join(self.column) if self.column else '*'
            where = f" where {self.where}" if self.where else ''
            limit = f" limit {start}, {size or 9999999999999}" if (start or size) else ''
            order = ' order by ' + ', '.join([x if y else f"{x} desc" for x, y in sort]) if sort else ''
            # sort = [(作者, True), (字数, True), ...], True表示升序
            conn, cursor, sheet = self.__get_conn()
            sql = f"select {sel} from {sheet}{where}{order}{limit}"
            cursor.execute(sql)
            kws = list(cursor.fetchall())
            conn.commit()
            cursor.close()
            if self.column: kws = [{key: kw.get(key) for key in self.column} for kw in kws]
            if rename:
                for kw in kws:
                    data = {new: kw.pop(old) for old,new in rename.items() if old in kw}
                    kw.update(data)
            if type_data == 'kw': kws = kws[0]
            return kws
        # 限定column
        elif type_ in (str, tuple):  # 输入多个字符串, 用逗号隔开, Python会自动打包成tuple
            column = {key} if (type_ is str) else set(key)
            return self.__copy(column=column)
        # Filter
        elif type_ is Filter:
            w1 = self.where
            w2 = key.a4173c5c8a87db003
            if w1 and w2:
                w = w1 if (w1 == w2) else f"({w1}) and ({w2})"
            else:
                w = w1 or w2 or ''
            return self.__copy(where=w)
        else:
            return None

    def delete_one(self):
        conn, cursor, sheet = self.__get_conn()
        where = f" where {self.where}" if self.where else ''
        sql = f"delete from {sheet}{where} limit 1"
        try:
            cursor.execute(sql)
            conn.commit()
            cursor.close()
            return cursor
        except Exception as err:
            conn.rollback()
            raise err

    def delete_all(self):
        conn, cursor, sheet = self.__get_conn()
        where = f" where {self.where}" if self.where else ''
        sql = f"delete from {sheet}{where}"
        try:
            cursor.execute(sql)
            conn.commit()
            cursor.close()
            return cursor
        except Exception as err:
            conn.rollback()
            raise err

    def reset_one(self, kw):
        if self.column:
            conn, cursor, sheet = self.__get_conn()
            kw = deepcopy(kw or {})
            kw = {k: kw.get(k) for k in self.column}
            kw = ','.join([f"{k}={jdumps(v)}" for k, v in kw.items()])
            where = f" where {self.where}" if self.where else ''
            sql = f"update {sheet} set {kw}{where} limit 1"
            try:
                cursor.execute(sql)
                conn.commit()
                cursor.close()
                return cursor
            except Exception as err:
                conn.rollback()
                raise err
        else:
            raise Exception('Column must be specified')

    def reset_all(self, kw):
        if self.column:
            conn, cursor, sheet = self.__get_conn()
            kw = deepcopy(kw or {})
            kw = {k: kw.get(k) for k in self.column}
            kw = ','.join([f"{k}={jdumps(v)}" for k, v in kw.items()])
            where = f" where {self.where}" if self.where else ''
            sql = f"update {sheet} set {kw}{where}"
            try:
                cursor.execute(sql)
                conn.commit()
                cursor.close()
                return cursor
            except Exception as err:
                conn.rollback()
                raise err
        else:
            raise Exception('Column must be specified')

    def update_one(self, kw):
        kw = deepcopy(kw or {})
        if self.column: kw = {k:v for k,v in kw.items() if k in self.column}
        if kw:
            conn, cursor, sheet = self.__get_conn()
            kw = ','.join([f"{k}={jdumps(v)}" for k, v in kw.items()])
            where = f" where {self.where}" if self.where else ''
            sql = f"update {sheet} set {kw}{where} limit 1"
            try:
                cursor.execute(sql)
                conn.commit()
                cursor.close()
                return cursor
            except Exception as err:
                conn.rollback()
                raise err
        else:
            return True

    def update_all(self, kw):
        kw = deepcopy(kw or {})
        if self.column: kw = {k:v for k,v in kw.items() if k in self.column}
        if kw:
            conn, cursor, sheet = self.__get_conn()
            kw = ','.join([f"{k}={jdumps(v)}" for k, v in kw.items()])
            where = f" where {self.where}" if self.where else ''
            sql = f"update {sheet} set {kw}{where}"
            try:
                cursor.execute(sql)
                conn.commit()
                cursor.close()
                return cursor
            except Exception as err:
                conn.rollback()
                raise err
        else:
            return True
    
    def sql(self, sql):
        conn, cursor, sheet = self.__get_conn()
        sql = sql.replace('{sheet}', sheet)
        try:
            cursor.execute(sql)
            r = list(cursor.fetchall())
            conn.commit()
            cursor.close()
            return r
        except Exception as err:
            conn.rollback()
            raise err

    def release(self):
        self.conn_pool.append(self)


class mpool():

    def __init__(self, address, sheet, minsize=10, maxsize=50):
        self.address = address
        self.sheet = sheet
        self.minsize = minsize
        self.maxsize = maxsize
        self.conn_pool = deque()
        for i in range(maxsize - minsize):
            self.conn_pool.append(None)
        for i in range(minsize):
            self.conn_pool.append(self.mkorm())

    def mkorm(self):
        return msheet(
            address=self.address, sheet=self.sheet,
            conn_pool=self.conn_pool, active=True
        )

    async def get(self):
        while True:
            try:
                orm = self.conn_pool.pop()
                if orm is None:
                    return self.mkorm()
                return orm
            except:
                await asyncio.sleep(1)


mc = Filter()  # 表示: mysql_column


class mo():  # 表示: mysql_oper
    re = re
    isin = isin
    notin = notin
